%% arguments:
%%
%% x0 - initial guess, primarily to ensure that there is at least one
%% feasible solution in the initial population. Actually, it's not clear
%% whether this is necessary...
%%
%% a, b - lower and upper bounds on x to provide a suitable domain for
%% the mutation operator and for creation of random individuals for the
%% generation of the initial population.
%%
%% It is assumed that the objective function is z=f(x) and z is
%% dimension >=1.
%%
%% The function returns a set of points which are meant to be an
%% approximate to the pareto front for the multi-objective function.
%% This set is represented by two arrays, one for the actual points and
%% one for the objective function values for these points. The total
%% number of function evaluations is also returned in case this is of
%% interest to the invoker.

function [paretox, paretoz, nfun] = moga (x0, a, b, f,
                                          %% optional variables follow
                                          ngarg,                              % number of generations, ng > 0
                                          psarg,                              % population size, ps > 0
                                          crarg,                              % crossover rate, cr in [0,1]
                                          mrarg,                              % mutation rate, mr in [0,1]
                                          elitearg,                           % number of elite members of population, elite > 0
                                          tsarg)                              % tournament size, ts > 1
  moga_initialise ();
  %% link up global variables used by the objective function
  %% evaluation procedure
  global No_evaluation
  global No_feasible
  global No_outrange
  No_evaluation = 0;
  No_feasible = 0;
  No_outrange = 0;

  %% the genetic algorithm that forms the basis of this method uses
  %% random numbers. the Matlab random number method, rand, always
  %% generates the same sequence of 'random' numbers if it is not
  %% initialised with a seed. we use the clock (i.e. the time) to
  %% generate a random seed for the random number generator to get
  %% different behaviour each time we invoke this method.
  rand ('state', sum (100*clock) );

  %% efficiency considerations
  %% ignore_function_time_stamp = 'all';

  %% default values for arguments
  cr = 0.7                              % crossover rate
  mr = 0.1                              % mutation rate
  n =  40                             % population size
  ne = n/2                            % elite size (half population) for when pareto front dominates
  ng = 50                             % number of generations
  ts = 2                                % tournament size
  nfun = 0;                           % number of function evaluations

  if nargin > 4, ng = ngarg; end;
  if nargin > 5, n = psarg; ne = n/2; end;
  if nargin > 6, cr = crarg; end;
  if nargin > 7, mr = mrarg; end;
  if nargin > 8, ne = elitearg; end;
  if nargin > 9, ts = tsarg; end;

  fprintf('MOGA parameters: \n\
       number of generations = %5d\n\
             population size = %5d\n\
              crossover rate = %8.2f\n\
               mutation rate = %8.2f\n\
     number of elite members = %5d\n\
             tournament size = %5d\n', ng, n, cr, mr, ne, ts)

  %% evaluate an actual point so that we can extract necessary size
  %% information (dimension of design space and number of objective
  %% function values in particular).
  z0 = f (x0);                 % evaluate an actual point
  d = size (x0, 1);            % dimension; we assume column vector...
  if d == 1
    error ('x0, the initial guess, should be a column vector.');
  end
  m = size (z0, 1);                   % number of objective function values

  fprintf('MOGA problem dimension %d with %d objectives.\n', d, m)

  %% create the initial population. the population data structure
  %% contains an array for the design points (each of which has
  %% dimension d), an array for the objective function values (size m
  %% for each point) and an array for the fitness values. Remember that
  %% we wish to minimise function evaluations as they are very
  %% expensive.

  pop.x = zeros (d, n);               % actual points in dimension d space
  pop.z = zeros (m, n);               % the m objective function values
  pop.f = zeros (n);                  % the fitness of the point
  pop.x (:,1) = x0;
  pop.z (:,1) = z0;
  for i = 2:n
    x = moga_createRandom (a,b);
    while ~ moga_diverse (x, pop)
      debugprint(1,'Random solution already in population... trying again.',x)
      x = moga_createRandom (a,b);
    end
    z = f (x);
    debugprint(1,'Adding solution to population.\n', [x;z])
    nfun = nfun + 1;
    pop.x (:, i) = x;
    pop.z (:, i) = z;
  end

  %% now enter a loop where we evaluate the fitness of all the members
  %% of the population, then do a number of selections with mating
  %% and/or mutation to create a new population.

  for g = 1:ng
    debugprint(0,'MOGA, generation', g)
    %% calculate the fitness of each of the members of the population
    [pop.f pareto] = moga_fitness (pop.x, pop.z);
    fprintf(stderr, '\r %8d %13.3g %13.3g ', g, pareto.z(1,1), pareto.z(end,end))
    debugprint(0, 'MOGA, best objective 1', [pareto.x(:,1);pareto.z(:,1)])
    debugprint(0, 'MOGA, best objective 2', [pareto.x(:,end);pareto.z(:,end)])
    if 0 == mod(g,ng/50)
      fprintf('%8d %13.3g %13.3g %13.3g %13.3g \n', g, pareto.z(:,1), pareto.z(:,end))
    end
    %% create a new population, starting with the pareto front
    %% (hoping that it is NOT the whole population)
    new = pareto;                   % form of elitism
    i = size (pareto.x, 2);         % number of entries already
    if i >= n
      %% as the pareto front is the whole of the population, we wouldn't have
      %% any room to add anything if we were to copy it all
      %% over. Instead, we copy over a number of solutions (NE)
      %% instead.
      disp (['Pareto set overwhelms population. Copying only some solutions into new population']);
      for j=1:ne
        [x z] = moga_select (pop,ts);
        new.x (:, j) = x;
        new.z (:, j) = z;
      end
      i = ne;
    end
    while i<n
      debugprint(2,'selection');
      %% select two members and then decide whether to mate and/or
      %% mutate these two members. we keep track (e1 and e2) of whether
      %% we need to re-evaluate either or both of the design points,
      %% remembering that we wish to avoid any evaluations if at all
      %% possible due to their very high computational cost.
      [x1 z1] = moga_select (pop,ts);
      [x2 z2] = moga_select (pop,ts);
      e1 = 0;                     % evaluation not required
      e2 = 0;
      if rand < cr                % do we mate? (crossover rate)
        [x1 x2] = moga_mate (x1, x2);
        debugprint(3,'crossover x1', x1)
        debugprint(3,'crossover x2', x2)
        %% disp (sprintf (' crossover x1=[%f %f %f %f] x2=[%f %f %f %f]\n', x1', x2'));
        e1 = 1;
        e2 = 1;
      end
      if rand < mr                % mutate first solution?
        x1 = moga_mutate (x1, a, b);
        debugprint(3,' mutated x1', x1);
        e1 = 1;
      end
      if moga_diverse (x1, new)   % only consider adding x1 if this preserves/enhances diversity
        if e1
          z1 = f (x1);
          nfun = nfun + 1;;
        end
        debugprint(3,'added x1 & z to population', [x1;z1])
        i = i+1;
        new.x (:, i) = x1;
        new.z (:, i) = z1;
      end
      if rand < mr                % mutate second?
        x2 = moga_mutate (x2, a, b);
        debugprint(3,' mutated x2', x2);
        e2 = 1;
      end
      if moga_diverse (x2, new)   % diverse?
        if e2
          z2 = f (x2);
          nfun = nfun + 1;
        end
        i = i+1;
        new.x (:, i) = x2;
        new.z (:, i) = z2;
        debugprint(3,' added x2 & z to pop', [x2; z2])
      end
    end
    %% make current population be the one we have just created
    pop = new;
  end
  paretox = pareto.x;
  paretoz = pareto.z;
  debugprint(1,'moga: at end, nfun=', nfun);

  %% Local Variables:
  %% time-stamp-line-limit: 1000
  %% time-stamp-format: '%04y.%02m.%02d %02H:%02M:%02S'
  %% time-stamp-active: t
  %% time-stamp-start: 'version = \''
  %% time-stamp-end: '\';'
  %% End:
